/*
    This header file is part of CadCam.

    This header file is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This header file is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CadCam.  If not, see <https://www.gnu.org/licenses/>.

    AUTOR     : Grotius
    DATE      : June 2020
    REVISION  : 7.6.2020     Created class
                9.6.2020     Added area & contour swap
    TODO      : 1. Test G5 gcode splines
    USAGE     : To use this header only library/file, include also structs.h to your project.
    PROTOCOL  : 1. Minimize the use of QT libraries to stay compatible with other program's.
    NOTES     : Control your gcode output with : https://ncviewer.com/
*/

#include <iostream>
#include <algorithm>
#include <structs.h>
#include <math.h>
#include <chrono> //high_resolution_clock
#include <thread>
#include <fstream>
#include <unistd.h> //user for find current directory


std::chrono::duration<double> elapsed;

extern double feedrate;
extern double cutheight;
extern double rapidheight;
extern double pierceheight;
extern double piercespeed;
extern double piercedelay;
extern double spline_resolution; //value >0 && <1
extern bool spline_G5;
extern std::string gcode_filename;

void area();
double sum_plus_controlpoints(int i);
void map_cw_ccw();
void swap_contour(std::vector<CONTOUR>::iterator it);
void process_gcode();

int G2(int i);
void gcode(int i);
void write(std::string text);
void intro();
void outtro();
void M3(int );
void M5();
std::vector<POINT> spline_as_linestrip(std::vector<std::vector<double>> C);
std::string get_current_dir();

void process_gcode(){
    auto start = std::chrono::high_resolution_clock::now();

    std::ofstream myfile;
    myfile.open(get_current_dir()+"/"+gcode_filename /*, std::ios::app*/); //ios app = append true
    myfile.clear();

    intro();
    //print contour sequence, based on keep parts together algoritme
    for(int i=0; i<sequence.size(); i++){
        //std::cout<<"contour: "<<sequence.at(i)<<std::endl;

        for(int j=0; j<contours.at(sequence.at(i)).sequence.size(); j++){
            int object=contours.at(sequence.at(i)).sequence.at(j);

            if(object==contours.at(sequence.at(i)).sequence.front()){
                M3(object);
            }

            gcode(object);

            if(object==contours.at(sequence.at(i)).sequence.size()){
                M5();
            }
        }
    }
    M5();
    outtro();

    myfile.close();

    auto finish = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> elapsed = finish - start;
    std::cout<<""<<std::endl;
    std::cout<<"time: "<<elapsed.count()<<" sec"<<std::endl;
    std::cout<<"Gcode done!"<<std::endl;
}

void gcode(int i){

    //line section
    if(objects.at(i).type=="line"){
        write("G1 X"+std::to_string(objects.at(i).end.x)+ " Y"+std::to_string(objects.at(i).end.y)+ " Z"+std::to_string(cutheight));
    }

    //circle section
    //I=offset xcenter-xstart, J=offset ycenter-ystart, G2=clockwise (cw), G3=counterclockwise (ccw)
    if(objects.at(i).type=="circle"){ //circle is single contour, is always at vector front position
        if(objects.at(i).red==255 && objects.at(i).green==0 && objects.at(i).blue==0 && objects.at(i).alpha==255){
            write("G2 I"+std::to_string(objects.at(i).center.x-objects.at(i).start.x)+" J"+std::to_string(objects.at(i).center.y-objects.at(i).start.y)+" F"+std::to_string(feedrate));
        } else { //color must be yellow, ccw
            write("G3 I"+std::to_string(objects.at(i).center.x-objects.at(i).start.x)+" J"+std::to_string(objects.at(i).center.y-objects.at(i).start.y)+" F"+std::to_string(feedrate));
        }
    }

    //arc section
    //X=xend, Y=yend. For arc given a G0 startposition and a XY endposition. http://linuxcnc.org/docs/html/gcode/g-code.html#gcode:g2-g3
    //I=offset xcenter-xstart, J=offset ycenter-ystart, G2=clockwise (cw), G3=counterclockwise (ccw)
    if(objects.at(i).type=="arc"){
        if(G2(i)==2){
            write("G2 X"+std::to_string(objects.at(i).end.x)+" Y"+std::to_string(objects.at(i).end.y)+" I"+std::to_string(objects.at(i).center.x-objects.at(i).start.x)+" J"+std::to_string(objects.at(i).center.y-objects.at(i).start.y)+" F"+std::to_string(feedrate));
        }
        if(G2(i)==3){
            write("G3 X"+std::to_string(objects.at(i).end.x)+" Y"+std::to_string(objects.at(i).end.y)+" I"+std::to_string(objects.at(i).center.x-objects.at(i).start.x)+" J"+std::to_string(objects.at(i).center.y-objects.at(i).start.y)+" F"+std::to_string(feedrate));
        }
    }

    if(objects.at(i).type=="spline" && spline_G5==false){ //write spline gcode as a linestrip
        std::vector<std::vector<double>> C; //2d array
        std::vector<POINT> points;
        std::vector<POINT>::iterator it;
        C.push_back({objects.at(i).start.x, objects.at(i).start.y}); //spline startpoint
        for(int j=0; j<objects.at(i).control.size(); j++){
            C.push_back({objects.at(i).control.at(j).x, objects.at(i).control.at(j).y}); //spline controlpoints
        }
        C.push_back({objects.at(i).end.x, objects.at(i).end.y}); //spline endpoint
        points=spline_as_linestrip(C);
        for(it=points.begin()+1; it<points.end(); it++){ //exclude the spline startpoint, begin+1
              write("G1 X"+std::to_string(it->x)+ " Y"+std::to_string(it->y)+ " Z"+std::to_string(cutheight));
        }
        //added....
        write("G1 X"+std::to_string(objects.at(i).end.x)+ " Y"+std::to_string(objects.at(i).end.y)+ " Z"+std::to_string(cutheight));
    }

    //spline section http://linuxcnc.org/docs/html/gcode/g-code.html#gcode:g5
}

void intro(){
    write("G21      (unit:mm)");
    write("G40      (cutter compensation off)");
    write("G90      (absolute distance, no offsets)");
    write("G64P0.01 (path following accuracy)");
    write("F1       (initialize feedrate)");
    write("G0 Z"+std::to_string(rapidheight));
}

void outtro(){
    write("M30 (program end)");
    write("(with great power comes great responsibility)");
}

void M3(int i){
    write("G0 X"+std::to_string(objects.at(i).start.x)+ " Y"+std::to_string(objects.at(i).start.y)+ " Z"+std::to_string(rapidheight));
    write("G0 Z"+std::to_string(pierceheight)+" (at pierceheight)");
    write("M3 (torch on)");
    write("G4 P"+std::to_string(piercedelay)+" (pierce delay)");
    write("G1 Z"+std::to_string(cutheight)+" (cutheight)");
}

void M5(){
    write("M5 (torch off)");
    write("G0 Z"+std::to_string(rapidheight));
}

void write(std::string text){
    //std::cout<<text<<std::endl;
    std::ofstream myfile;
    myfile.open(get_current_dir()+"/"+gcode_filename, std::ios::app);
    myfile << text << std::endl;
}

int G2(int i){
    //formula if arc is cw or ccw.. this is nice..
    //To determine which side of the line from A=(x1,y1) to B=(x2,y2) a point P=(x,y) falls on you need to compute the value:-
    //d=(x−x1)(y2−y1)−(y−y1)(x2−x1)
    //If d<0 then the point lies on one side of the line, and if d>0 then it lies on the other side. If d=0 then the point lies exactly line.
    //If d>0 we got a G2
    double x1=objects.at(i).start.x;
    double y1=objects.at(i).start.y;
    double x2=objects.at(i).end.x;
    double y2=objects.at(i).end.y;
    double x=objects.at(i).center.x;
    double y=objects.at(i).center.y;
    double d=((x-x1)*(y2-y1))-((y-y1)*(x2-x1));
    if(d>0){return 2;}
    if(d<0){return 3;}

}

void area(){
    std::vector<CONTOUR>::iterator it;
    std::vector<int>::iterator it1;
    for(it=contours.begin(); it<contours.end(); it++){
        double sum=0;
        if(it->contourtype!=0){  //0=open, 1=closed, 2=single object closed
            for(it1=it->sequence.begin();  it1<it->sequence.end(); it1++){
                sum+=sum_plus_controlpoints(*it1);
            }
        }
        it->area=sum/2;
        //std::cout<<"area: "<<sum/2<<std::endl;
    }
}

double sum_plus_controlpoints(int i){
    double sum=0;
    //if object has controlpoints
    if(objects.at(i).control.size()>0){
        sum+= (objects.at(i).control.front().x-objects.at(i).start.x) * (objects.at(i).control.front().y+objects.at(i).start.y);
        for(int j=0; j<objects.at(i).control.size()-1; j++){
            sum+= (objects.at(i).control.at(j+1).x-objects.at(i).control.at(j).x) * (objects.at(i).control.at(j+1).y+objects.at(i).control.at(j).y);
        }
        sum+= (objects.at(i).end.x-objects.at(i).control.back().x) * (objects.at(i).end.y+objects.at(i).control.back().y);
    }

    //if object has no controlpoints
    if(objects.at(i).control.size()==0){
        sum+= (objects.at(i).end.x-objects.at(i).start.x) * (objects.at(i).end.y+objects.at(i).start.y);
    }
    return sum;
}

void map_cw_ccw(){
    std::vector<CONTOUR>::iterator it;
    for(it=contours.begin(); it<contours.end(); it++){
        if(it->area==0){
            //do nothing, open contour
        }
        if(it->area>0){
            //cw contour, check if contour color is red, if not, swap contour
            int object=it->sequence.front();

            if(objects.at(object).red==255 && objects.at(object).green==0 && objects.at(object).blue==0 && objects.at(object).alpha==255){
                //std::cout<<"cw contour, color is red, ok"<<std::endl;
            }
            if(objects.at(object).red==255 && objects.at(object).green==255 && objects.at(object).blue==0 && objects.at(object).alpha==255){
                //std::cout<<"cw contour, color is yellow, not ok, swap"<<std::endl;
                swap_contour(it);
            }
        }
        if(it->area<0){
            int object=it->sequence.front();

            if(objects.at(object).red==255 && objects.at(object).green==0 && objects.at(object).blue==0 && objects.at(object).alpha==255){
                //std::cout<<"ccw contour, color is red, not ok, swap"<<std::endl;
                swap_contour(it);
            }
            if(objects.at(object).red==255 && objects.at(object).green==255 && objects.at(object).blue==0 && objects.at(object).alpha==255){
                //std::cout<<"ccw contour, color is yellow, ok"<<std::endl;
            }
        }
    }
}
void swap_contour(std::vector<CONTOUR>::iterator it){
    std::vector<int>::iterator it1;
    std::reverse(it->sequence.begin(), it->sequence.end());
    for(it1=it->sequence.begin(); it1<it->sequence.end(); it1++){
        //swap it1 parameters
        double start_x = objects.at(*it1).start.x;
        double start_y = objects.at(*it1).start.y;
        double start_z = objects.at(*it1).start.z;
        objects.at(*it1).start.x=objects.at(*it1).end.x;
        objects.at(*it1).start.y=objects.at(*it1).end.y;
        objects.at(*it1).start.z=objects.at(*it1).end.z;
        objects.at(*it1).end.x=start_x;
        objects.at(*it1).end.y=start_y;
        objects.at(*it1).end.z=start_z;
        std::reverse(objects.at(*it1).control.begin(), objects.at(*it1).control.end());
    }
}

std::vector<POINT> spline_as_linestrip(std::vector<std::vector<double>> C){

    POINT point;
    std::vector<POINT> points;

    double S[C.size()-1][2][4];
    double w, b[C.size()], d[C.size()], x[C.size()];
    int i, dim;
    int n = C.size()-1; // number of splines

    for(dim=0; dim<2; dim++)
    {
        // define d[]:
        d[0]=3.0*(C[1][dim]-C[0][dim]);
        for(i=1; i<n; i++)
            d[i]=3.0*(C[i+1][dim]-C[i-1][dim]);
        d[n]=3.0*(C[n][dim]-C[n-1][dim]);
        // forward sweep: (simplified tridiagonal solution: all a[i]=1 and all c[i]=1)
        b[0]=2.0;
        for(i=1; i<C.size(); i++)
        {
            w = 1.0/b[i-1];
            b[i] = 4.0 - w;
            d[i] -= (w*d[i-1]);
        }
        // calculate solution vector x[i] = D[i]:
        // (Wikipedia x[] = Wolfram D[])
        x[n]=d[n]/b[n];
        for(i=n-1; i>=0; i--)
            x[i]=(d[i]-x[i+1])/b[i];
        // calculate spline S[i][dim] a, b, c and d:
        for(i=0; i<n; i++)
        {
            S[i][dim][0]=C[i][dim];
            S[i][dim][1]=x[i];
            S[i][dim][2]=3.0*(C[i+1][dim]-C[i][dim]) - 2.0*x[i] - x[i+1];
            S[i][dim][3]=2.0*(C[i][dim]-C[i+1][dim]) + x[i] + x[i+1];
        }
    }

    //glBegin(GL_LINE_STRIP);
    for(int p=0; p<C.size()-1; p++)  //spline points
    {
        for(double t=0; t<1; t+=spline_resolution)  //time 0-1 for each spline
        {
            point.x = ((S[p][0][3]*t+S[p][0][2])*t+S[p][0][1])*t+S[p][0][0];
            point.y = ((S[p][1][3]*t+S[p][1][2])*t+S[p][1][1])*t+S[p][1][0];
            //std::cout<< "x:" << px << " y:" << py << std::endl;
            // glVertex2d(px,py);
            points.push_back(point);
        }
    }
    return points;
}

std::string get_current_dir() {
   char buff[FILENAME_MAX]; //create string buffer to hold path
   getcwd( buff, FILENAME_MAX );
   std::string current_working_dir(buff);
   return current_working_dir;
}

























